clear
close all

maxits = 250;
f = double(imread('boat.png'))/255;
%f = double(imread('Eagle.jpg'))/255;

shift = 6;      % number of shifted input images

scale = 0.5;
sig = 1.;       % standard deviation of Gaussian Convolution
g_patchRadius = 1;

lmd = 0.1;      % weight of regularizer
tau = 0.01;     % gradient descent step size
eps = 1e-6;     % epsilon for total-variation discontinuity
check = 10;     % how often to update visualization

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%
% Creating input images 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% convolving image with a gaussian kernel with standard deviation sig
filter = fspecial('gaussian',[g_patchRadius,g_patchRadius],sig);
f = imfilter(f,filter,'replicate');

% calculating dimensions
[H, W] = size(f);
if not (mod(H,2) == 0 && mod(W,2) == 0)
    % make sure image dimensions are divisible by 2
    f = imresize(f, [H - mod(H,2), W - mod(W,2)]);
    [H, W] = size(f);
end 
W = W - shift;
n = W*H;
w = scale*W;
h = scale*H;

% generate input frames by shifting each by 1 pixel
g = zeros(h, w, shift);
for i = 1:shift
   g(:,:,i) = imresize(f(:,i+1:end-(shift-i)), scale, 'method', 'nearest');
end

g_vec = zeros(w*h, shift);
for i = 1:shift
    g_vec(:,i) = reshape(g(:,:,i), w*h, 1);
end


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Building Upsampling Operator U
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

one_vec=ones(2,1);
I_u1 = speye(h);
I_u2 = speye(w);
u = kron(I_u1, one_vec);
u = [u;u];
U = kron(I_u2, u);

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct Gaussian Blurring Matrix B 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Gaussian 1d kernel
filter = fspecial('gaussian',[1, 2*g_patchRadius+1], sig);

G_y  =spdiags(repmat(filter, H, 1), -1:1, H, H);
G_y = kron(speye(W, W), G_y);

G_x = spdiags(repmat(filter, W, 1), -1:1, W, W);
G_x = kron(G_x', speye(H, H));

% implement Gaussian blur as concatenation of x and y blur
B = G_x * G_y;

%% %%%%%%%%%%%%%%%%%%%%%%%%%%
% Construct warping matrix S 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

S = cell(1,shift);
for i=1:shift
    S{i} = spdiags(ones(W*H,1), i*H, W*H, W*H);
end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Build averaging matrix A 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

factor = 2;
Avg_rows = kron(speye(H*W/factor), ones(factor)/factor);
Avg_cols = kron(speye(W/factor), kron(ones(factor)/factor, speye(H)));
A = Avg_rows * Avg_cols;

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Gradient and Divergence Operator 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[grad, Dx, Dy] = spmat_gradient2d(H, W, 1);
div = -grad';

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Calculate coefficient matrices M_i = A*B*S_i
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

M = cell(1,shift);
for i=1:shift
    M{i} = (A*B*S{i});
end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Upsample frame images 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

g_vec_u = zeros(W*H, shift);
for i = 1:shift
    g_vec_u(:,i) = U*g_vec(:,i);
end

u = g_vec_u(:,1);

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Gradient descent
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for it=1:maxits
    
    update = zeros(size(u));
    for i=1:shift
        update = update+M{i}'*(M{i}*u-g_vec_u(:,i));     
    end
    ux = Dx*u;
    uy = Dy*u;
    abs_uxy = sqrt(ux.^2 + uy.^2 + eps);
    ux_ = ux./abs_uxy;
    uy_ = uy./abs_uxy;
    update = 2*update - lmd*div*[ux_;uy_];
    u = u - tau*update;

    if (mod(it, check) == 0)       
        fprintf('Iteration: %d\n',it);
        % reshape vector u to a full image
        res = reshape(u,H,W);
        display(res, g);
        drawnow;
    end
    
end

% display final result
res = reshape(u,H,W);
display(res, g);


function display(res, g)
    subplot(1,2,1)
    imshow(g(:,:,end))
    title('One Input Image');

    subplot(1,2,2)
    imshow(res)
    title('Super Resolved Image');
end
